#!/usr/bin/env python

import sys
import os
import socket
import argparse
import subprocess
import signal
import fcntl
import stat

criu_bin = '../../criu/criu'

sk_type_s = {
    socket.SOCK_STREAM: "S",
    socket.SOCK_DGRAM: "D",
}

# Actions that can be done by test. Actions are not only syscall
# names to call, but also arguments with which to do it
#
# Each action consists of
# - arguments, e.g. type of socket, or socket id to work on
# - act() method which just generates an record
# - do() method, that actually does what's required
# - show() method to return the string description of what's done


def mk_socket(st, typ):
    st.sk_id += 1
    sk = sock(st.sk_id, typ)
    st.add_socket(sk)
    return sk


class act_socket:
    def __init__(self, typ):
        self.typ = typ

    def act(self, st):
        sk = mk_socket(st, self.typ)
        self.sk_id = sk.sk_id

    def do(self, st):
        sk = socket.socket(socket.AF_UNIX, self.typ, 0)
        st.real_sockets[self.sk_id] = sk

    def show(self):
        return 'socket(%s) = %d' % (sk_type_s[self.typ], self.sk_id)


class act_close:
    def __init__(self, sk_id):
        self.sk_id = sk_id

    def act(self, st):
        sk = st.get_socket(self.sk_id)
        st.del_socket(sk)
        for ic in sk.icons:
            sk = st.get_socket(ic)
            st.del_socket(sk)

    def do(self, st):
        sk = st.real_sockets.pop(self.sk_id)
        sk.close()

    def show(self):
        return 'close(%d)' % self.sk_id


class act_listen:
    def __init__(self, sk_id):
        self.sk_id = sk_id

    def act(self, st):
        sk = st.get_socket(self.sk_id)
        sk.listen = True

    def do(self, st):
        sk = st.real_sockets[self.sk_id]
        sk.listen(10)

    def show(self):
        return 'listen(%d)' % self.sk_id


class act_bind:
    def __init__(self, sk_id, name_id):
        self.sk_id = sk_id
        self.name_id = name_id

    def act(self, st):
        sk = st.get_socket(self.sk_id)
        sk.name = self.name_id

    def do(self, st):
        sk = st.real_sockets[self.sk_id]
        sk.bind(sock.real_name_for(self.name_id))

    def show(self):
        return 'bind(%d, $name-%d)' % (self.sk_id, self.name_id)


class act_connect:
    def __init__(self, sk_id, listen_sk_id):
        self.sk_id = sk_id
        self.lsk_id = listen_sk_id

    def act(self, st):
        sk = st.get_socket(self.sk_id)
        if st.sk_type == socket.SOCK_STREAM:
            lsk = st.get_socket(self.lsk_id)
            psk = mk_socket(st, socket.SOCK_STREAM)
            psk.visible = False
            sk.peer = psk.sk_id
            psk.peer = sk.sk_id
            psk.name = lsk.name
            lsk.icons.append(psk.sk_id)
            lsk.icons_seq += 1
        else:
            sk.peer = self.lsk_id
            psk = st.get_socket(self.lsk_id)
            psk.icons_seq += 1

    def do(self, st):
        sk = st.real_sockets[self.sk_id]
        sk.connect(sock.real_name_for(self.lsk_id))

    def show(self):
        return 'connect(%d, $name-%d)' % (self.sk_id, self.lsk_id)


class act_accept:
    def __init__(self, sk_id):
        self.sk_id = sk_id

    def act(self, st):
        lsk = st.get_socket(self.sk_id)
        iid = lsk.icons.pop(0)
        nsk = st.get_socket(iid)
        nsk.visible = True
        self.nsk_id = nsk.sk_id

    def do(self, st):
        sk = st.real_sockets[self.sk_id]
        nsk, ai = sk.accept()
        if self.nsk_id in st.real_sockets:
            raise Exception("SK ID conflict")
        st.real_sockets[self.nsk_id] = nsk

    def show(self):
        return 'accept(%d) = %d' % (self.sk_id, self.nsk_id)


class act_sendmsg:
    def __init__(self, sk_id, to_id):
        self.sk_id = sk_id
        self.to_id = to_id
        self.direct_send = None

    def act(self, st):
        sk = st.get_socket(self.sk_id)
        msg = (sk.sk_id, sk.outseq)
        self.msg_id = sk.outseq
        sk.outseq += 1
        psk = st.get_socket(self.to_id)
        psk.inqueue.append(msg)
        self.direct_send = (sk.peer == psk.sk_id)

    def do(self, st):
        sk = st.real_sockets[self.sk_id]
        msgv = act_sendmsg.msgval(self.msg_id)
        if self.direct_send:
            sk.send(msgv)
        else:
            sk.sendto(msgv, sock.real_name_for(self.to_id))

    def show(self):
        return 'send(%d, %d, $message-%d)' % (self.sk_id, self.to_id,
                                              self.msg_id)

    @staticmethod
    def msgval(msgid, pref=''):
        return '%sMSG%d' % (pref, msgid)


#
# Description of a socket
#
class sock:
    def __init__(self, sk_id, sock_type):
        # ID of a socket. Since states and sockets are cloned
        # while we scan the tree of states the only valid way
        # to address a socket is to find one by ID.
        self.sk_id = sk_id
        # The socket.SOCK_FOO value
        self.sk_type = sock_type
        # Sockets that haven't yet been accept()-ed are in the
        # state, but user cannot operate on them. Also this
        # invisibility contributes to state description since
        # connection to not accepted socket is not the same
        # as connection to accepted one.
        self.visible = True
        # The listen() was called.
        self.listen = False
        # The bind() was called. Also set by accept(), the name
        # inherits from listener.
        self.name = None
        # The connect() was called. Set on two sockets when the
        # connect() is called.
        self.peer = None
        # Progress on accepting connections. Used to check when
        # it's OK to close the socket (see comment below).
        self.icons_seq = 0
        # List of IDs of sockets that can be accept()-ed
        self.icons = []
        # Number to generate message contents.
        self.outseq = 0
        # Incoming queue of messages.
        self.inqueue = []

    def clone(self):
        sk = sock(self.sk_id, self.sk_type)
        sk.visible = self.visible
        sk.listen = self.listen
        sk.name = self.name
        sk.peer = self.peer
        sk.icons_seq = self.icons_seq
        sk.icons = list(self.icons)
        sk.outseq = self.outseq
        sk.inqueue = list(self.inqueue)
        return sk

    def get_actions(self, st):
        if not self.visible:
            return []

        if st.sk_type == socket.SOCK_STREAM:
            return self.get_stream_actions(st)
        else:
            return self.get_dgram_actions(st)

    def get_send_action(self, to, st):
        # However, if peer has a message from us at
        # the queue tail, sending a new one doesn't
        # really make sense
        want_msg = True
        if len(to.inqueue) != 0:
            lmsg = to.inqueue[-1]
            if lmsg[0] == self.sk_id:
                want_msg = False
        if want_msg:
            return [act_sendmsg(self.sk_id, to.sk_id)]
        else:
            return []

    def get_stream_actions(self, st):
        act_list = []

        # Any socket can be closed, but closing a socket
        # that hasn't contributed to some new states is
        # just waste of time, so we close only connected
        # sockets or listeners that has at least one
        # incoming connection pendig or served

        if self.listen:
            if self.icons:
                act_list.append(act_accept(self.sk_id))
            if self.icons_seq:
                act_list.append(act_close(self.sk_id))
        elif self.peer:
            act_list.append(act_close(self.sk_id))
            # Connected sockets can send and receive messages
            # But receiving seem not to produce any new states,
            # so only sending
            # Also sending to a closed socket doesn't work
            psk = st.get_socket(self.peer, True)
            if psk:
                act_list += self.get_send_action(psk, st)
        else:
            for psk in st.sockets:
                if psk.listen and psk.name:
                    act_list.append(act_connect(self.sk_id, psk.sk_id))

            # Listen on not-bound socket is prohibited as
            # well as binding a listening socket
            if not self.name:
                # TODO: support for file paths (see real_name_for)
                # TODO: these names can overlap each other
                act_list.append(act_bind(self.sk_id, self.sk_id))
            else:
                act_list.append(act_listen(self.sk_id))

        return act_list

    def get_dgram_actions(self, st):
        act_list = []

        # Dgram socket can bind at any time
        if not self.name:
            act_list.append(act_bind(self.sk_id, self.sk_id))

        # Can connect to peer-less sockets
        for psk in st.sockets:
            if psk == self:
                continue
            if psk.peer != None and psk.peer != self.sk_id:
                # Peer by someone else, can do nothing
                continue

            # Peer-less psk or having us as peer
            # We can connect to or send messages
            if psk.name and self.peer != psk.sk_id:
                act_list.append(act_connect(self.sk_id, psk.sk_id))

            if psk.name or self.peer == psk.sk_id:
                act_list += self.get_send_action(psk, st)

        if self.outseq != 0 or self.icons_seq != 0:
            act_list.append(act_close(self.sk_id))

        return act_list

    @staticmethod
    def name_of(sk):
        if not sk:
            return 'X'
        elif not sk.visible:
            return 'H'
        elif sk.name:
            return 'B'
        else:
            return 'A'

    @staticmethod
    def real_name_for(sk_id):
        return "\0" + "CRSK%d" % sk_id

    # The describe() generates a string that represents
    # a state of a socket. Called by state.describe(), see
    # comment there about what description is.
    def describe(self, st):
        dsc = '%s' % sk_type_s[self.sk_type]
        dsc += sock.name_of(self)

        if self.listen:
            dsc += 'L'
        if self.peer:
            psk = st.get_socket(self.peer, True)
            dsc += '-C%s' % sock.name_of(psk)
        if self.icons:
            i_dsc = ''
            for c in self.icons:
                psk = st.get_socket(c)
                psk = st.get_socket(psk.peer, True)
                i_dsc += sock.name_of(psk)
            dsc += '-I%s' % i_dsc
        if self.inqueue:
            froms = set()
            for m in self.inqueue:
                froms.add(m[0])
            q_dsc = ''
            for f in froms:
                fsk = st.get_socket(f, True)
                q_dsc += sock.name_of(fsk)
            dsc += '-M%s' % q_dsc
        return dsc


class state:
    def __init__(self, max_sockets, sk_type):
        self.sockets = []
        self.sk_id = 0
        self.steps = []
        self.real_sockets = {}
        self.sockets_left = max_sockets
        self.sk_type = sk_type

    def add_socket(self, sk):
        self.sockets.append(sk)

    def del_socket(self, sk):
        self.sockets.remove(sk)

    def get_socket(self, sk_id, can_be_null=False):
        for sk in self.sockets:
            if sk.sk_id == sk_id:
                return sk

        if not can_be_null:
            raise Exception("%d socket not in list" % sk_id)

        return None

    def get_actions(self):
        act_list = []

        # Any socket in the state we can change it
        for sk in self.sockets:
            act_list += sk.get_actions(self)

        if self.sockets_left > 0:
            act_list.append(act_socket(self.sk_type))
            self.sockets_left -= 1

        return act_list

    def clone(self):
        nst = state(self.sockets_left, self.sk_type)
        for sk in self.sockets:
            nst.sockets.append(sk.clone())
        nst.sk_id = self.sk_id
        nst.steps = list(self.steps)
        return nst

    # Generates textual description of a state. Different states
    # may have same descriptions, e.g. if we have two sockets and
    # only one of them is in listen state, we don't care which
    # one in which. At the same time really different states
    # shouldn't map to the same string.
    def describe(self):
        sks = [x.describe(self) for x in self.sockets]
        sks = sorted(sks)
        return '_'.join(sks)


def set_nonblock(sk):
    fd = sk.fileno()
    flags = fcntl.fcntl(fd, fcntl.F_GETFL)
    fcntl.fcntl(fd, fcntl.F_SETFL, flags | os.O_NONBLOCK)


CHK_FAIL_UNKNOWN = 10
CHK_FAIL_SOCKET = 11
CHK_FAIL_STAT = 12
CHK_FAIL_LISTEN = 13
CHK_FAIL_NAME = 14
CHK_FAIL_ACCEPT = 15
CHK_FAIL_RECV_0 = 16
CHK_FAIL_RECV_MIX = 17
CHK_FAIL_CONNECT = 18
CHK_FAIL_CONNECT2 = 19
CHK_FAIL_KILLED = 20
CHK_FAIL_DUMP = 21
CHK_FAIL_RESTORE = 22

CHK_PASS = 42

fail_desc = {
    CHK_FAIL_UNKNOWN: 'Aliens invaded the test',
    CHK_FAIL_LISTEN: 'Listen state lost on restore',
    CHK_FAIL_NAME: 'Name lost on restore',
    CHK_FAIL_ACCEPT: 'Incoming connection lost on restore',
    CHK_FAIL_RECV_0: 'Message lost on restore',
    CHK_FAIL_RECV_MIX: 'Message misorder on restore',
    CHK_FAIL_CONNECT: 'Connectivity broken on restore',
    CHK_FAIL_CONNECT2: 'Connectivity broken the hard way on restore',
    CHK_FAIL_KILLED: 'Test process died unexpectedly',
    CHK_FAIL_DUMP: 'Cannot dump',
    CHK_FAIL_RESTORE: 'Cannot restore',
}


def chk_real_state(st):
    # Before enything else -- check that we still have
    # all the sockets at hands
    for sk in st.sockets:
        if not sk.visible:
            continue

        # In theory we can have key-not-found exception here,
        # but this has nothing to do with sockets restore,
        # since it's just bytes in memory, so ... we assume
        # that we have object here and just check for it in
        # the fdtable
        rsk = st.real_sockets[sk.sk_id]
        try:
            s_st = os.fstat(rsk.fileno())
        except:
            print('FAIL: Socket %d lost' % sk.sk_id)
            return CHK_FAIL_SOCKET
        if not stat.S_ISSOCK(s_st.st_mode):
            print('FAIL: Not a socket %d at %d' % (sk.sk_id, rsk.fileno()))
            return CHK_FAIL_STAT

    # First -- check the listen states and names
    for sk in st.sockets:
        if not sk.visible:
            continue

        rsk = st.real_sockets[sk.sk_id]
        r_listen = rsk.getsockopt(socket.SOL_SOCKET, socket.SO_ACCEPTCONN)
        if (sk.listen and r_listen == 0) or (not sk.listen and r_listen == 1):
            print("FAIL: Socket %d listen %d, expected %d" %
                  (sk.sk_id, r_listen, sk.listen and 1 or 0))
            return CHK_FAIL_LISTEN

        if sk.name:
            r_name = rsk.getsockname()
            w_name = sock.real_name_for(sk.name)
            if r_name != w_name:
                print('FAIL: Socket %d name mismatch [%s], want [%s]' %
                      (sk.sk_id, r_name, w_name))
                return CHK_FAIL_NAME

    # Second -- check (accept) pending connections
    for sk in st.sockets:
        if not sk.listen:
            continue

        rsk = st.real_sockets[sk.sk_id]
        set_nonblock(rsk)

        while sk.icons:
            # Do act_accept to change the state properly
            # and not write the code twice
            acc = act_accept(sk.sk_id)
            acc.act(st)
            try:
                acc.do(st)
            except:
                print('FAIL: Cannot accept pending connection for %d' %
                      sk.sk_id)
                return CHK_FAIL_ACCEPT

            print('  `- did %s' % acc.show())

    # Third -- check inqueues
    for sk in st.sockets:
        if not sk.inqueue:
            continue

        rsk = st.real_sockets[sk.sk_id]
        set_nonblock(rsk)

        while sk.inqueue:
            msg = sk.inqueue.pop(0)
            try:
                r_msg, m_from = rsk.recvfrom(128)
            except:
                print('FAIL: No message in queue for %d' % sk.sk_id)
                return CHK_FAIL_RECV_0

            w_msg = act_sendmsg.msgval(msg[1])
            if r_msg != w_msg:
                print('FAIL: Message misorder: %s want %s (from %d)' %
                      (r_msg, w_msg, msg[0]))
                return CHK_FAIL_RECV_MIX

            # TODO -- check sender
            print('  `- recvd %d.%d msg %s -> %d' %
                  (msg[0], msg[1], m_from, sk.sk_id))

    # Finally, after all sockets are visible and all inqueues are
    # drained -- check the sockets connectivity
    for sk in st.sockets:
        if not sk.peer:
            continue

        # Closed connection with one peer alive. Cannot check.
        if not sk.peer in st.real_sockets:
            continue

        rsk = st.real_sockets[sk.sk_id]
        psk = st.real_sockets[sk.peer]
        set_nonblock(psk)
        msgv = act_sendmsg.msgval(3 * sk.sk_id + 5 * sk.peer,
                                  'C')  # just random

        try:
            rsk.send(msgv)
            rmsg = psk.recv(128)
        except:
            print('FAIL: Connectivity %d -> %d lost' % (sk.sk_id, sk.peer))
            return CHK_FAIL_CONNECT

        # If sockets are not connected the recv above
        # would generate exception and the check would
        # fail. But just in case we've screwed the queues
        # the hard way -- also check for the message being
        # delivered for real
        if rmsg != msgv:
            print('FAIL: Connectivity %d -> %d not verified' %
                  (sk.sk_id, sk.peer))
            return CHK_FAIL_CONNECT2

        print('  `- checked %d -> %d with %s' % (sk.sk_id, sk.peer, msgv))

    return CHK_PASS


def chk_state(st, opts):
    print("Will check state")

    sigsk_name = "\0" + "CRSIGSKC"
    signal_sk = socket.socket(socket.AF_UNIX, socket.SOCK_DGRAM, 0)
    signal_sk.bind(sigsk_name)

    # FIXME Ideally call to criu should be performed by the run_state's
    # pid!=0 branch, but for simplicity we fork the kid which has the
    # same set of sockets we do, then dump it. Then restore and notify
    # via dgram socket to check its state. Current task still has all
    # the same sockets :) so we close them not to produce bind() name
    # conflicts on restore

    pid = os.fork()
    if pid == 0:
        msg = signal_sk.recv(64)
        ret = chk_real_state(st)
        sys.exit(ret)

    signal_sk.close()
    for rsk in st.real_sockets.values():
        rsk.close()

    print("`- dump")
    img_path = "sti_" + st.describe()
    try:
        os.mkdir(img_path)
        subprocess.check_call([
            criu_bin, "dump", "-t",
            "%d" % pid, "-D", img_path, "-v4", "-o", "dump.log", "-j"
        ])
    except:
        print("Dump failed")
        os.kill(pid, signal.SIGKILL)
        return CHK_FAIL_DUMP

    print("`- restore")
    try:
        os.waitpid(pid, 0)
        subprocess.check_call([
            criu_bin, "restore", "-D", img_path, "-v4", "-o", "rst.log", "-j",
            "-d", "-S"
        ])
    except:
        print("Restore failed")
        return CHK_FAIL_RESTORE

    print("`- check")
    signal_sk = socket.socket(socket.AF_UNIX, socket.SOCK_DGRAM, 0)
    try:
        signal_sk.sendto('check', sigsk_name)
    except:
        # Probably the peer has died before us or smth else went wrong
        os.kill(pid, signal.SIGKILL)

    wp, status = os.waitpid(pid, 0)
    if os.WIFEXITED(status):
        status = os.WEXITSTATUS(status)
        if status != CHK_PASS:
            print("`- exited with %d" % status)
            return status
    elif os.WIFSIGNALED(status):
        status = os.WTERMSIG(status)
        print("`- killed with %d" % status)
        return CHK_FAIL_KILLED
    else:
        return CHK_FAIL_UNKNOWN

    return CHK_PASS


def run_state(st, opts):
    print("Will run state")
    pid = os.fork()
    if pid != 0:
        wpid, status = os.wait()
        if os.WIFEXITED(status):
            status = os.WEXITSTATUS(status)
        elif os.WIFSIGNALED(status):
            status = CHK_FAIL_KILLED
        else:
            status = CHK_FAIL_UNKNOWN
        return status

    # Try the states in subprocess so that once
    # it exits the created sockets are removed
    for step in st.steps:
        step.do(st)

    if not opts.run:
        ret = chk_state(st, opts)
    else:
        ret = chk_real_state(st)

    sys.exit(ret)


def proceed(st, seen, failed, opts, depth=0):
    desc = st.describe()
    if not desc:
        pass
    elif not desc in seen:
        # When scanning the tree we run and try only states that
        # differ, but don't stop tree traversal on them. This is
        # because sometimes we can get into the already seen state
        # using less steps and it's better to proceed as we have
        # depth to move forward and generate more states.
        seen[desc] = len(st.steps)
        print('%s' % desc)
        for s in st.steps:
            print('\t%s' % s.show())

        if not opts.gen:
            ret = run_state(st, opts)
            if ret != CHK_PASS:
                failed.add((desc, ret))
                if not opts.keep:
                    return False
    else:
        # Don't even proceed with this state if we've already
        # seen one but get there with less steps
        seen_score = seen[desc]
        if len(st.steps) > seen_score:
            return True
        else:
            seen[desc] = len(st.steps)

    if depth >= opts.depth:
        return True

    actions = st.get_actions()
    for act in actions:
        nst = st.clone()
        act.act(nst)
        nst.steps.append(act)
        if not proceed(nst, seen, failed, opts, depth + 1):
            return False

    return True


p = argparse.ArgumentParser("CRIU test suite")
p.add_argument("--depth", help="Depth of generated tree", default='8')
p.add_argument("--sockets", help="Maximum number of sockets", default='1')
p.add_argument("--dgram", help="Use SOCK_DGRAM sockets", action='store_true')
p.add_argument("--stream", help="Use SOCK_STREAM sockets", action='store_true')
p.add_argument("--gen",
               help="Only generate and show states",
               action='store_true')
p.add_argument("--run",
               help="Run the states, but don't C/R",
               action='store_true')
p.add_argument("--keep", help="Don't stop on error", action='store_true')
opts = p.parse_args()
opts.depth = int(opts.depth)

# XXX: does it make any sense to mix two types in one go?
if opts.stream and opts.dgram:
    print('Choose only one type')
    sys.exit(1)

if opts.stream:
    sk_type = socket.SOCK_STREAM
elif opts.dgram:
    sk_type = socket.SOCK_DGRAM
else:
    print('Choose some type')
    sys.exit(1)

st = state(int(opts.sockets), sk_type)
seen = {}
failed = set()
proceed(st, seen, failed, opts)

if len(failed) == 0:
    print('PASS (%d states)' % len(seen))
else:
    print('FAIL %d/%d' % (len(failed), len(seen)))
    for f in failed:
        print("\t%-50s: %s" %
              (f[0], fail_desc.get(f[1], 'unknown reason %d' % f[1])))



#!/usr/bin/python

import sys
import rpc_pb2 as rpc

from setup_swrk import setup_swrk

print('Connecting to CRIU in swrk mode to check the version:')

swrk, s1 = setup_swrk()

# Create criu msg, set it's type to dump request
# and set dump options. Checkout more options in protobuf/rpc.proto
req = rpc.criu_req()
req.type = rpc.VERSION

# Send request
s1.send(req.SerializeToString())

# Recv response
resp = rpc.criu_resp()
MAX_MSG_SIZE = 1024
resp.ParseFromString(s1.recv(MAX_MSG_SIZE))

if resp.type != rpc.VERSION:
    print('RPC: Unexpected msg type')
    sys.exit(-1)
else:
    if resp.success:
        print('RPC: Success')
        print('CRIU major %d' % resp.version.major_number)
        print('CRIU minor %d' % resp.version.minor_number)
        if resp.version.HasField('gitid'):
            print('CRIU gitid %s' % resp.version.gitid)
        if resp.version.HasField('sublevel'):
            print('CRIU sublevel %s' % resp.version.sublevel)
        if resp.version.HasField('extra'):
            print('CRIU extra %s' % resp.version.extra)
        if resp.version.HasField('name'):
            print('CRIU name %s' % resp.version.name)
    else:
        print('Fail')
        sys.exit(-1)



import fcntl
import os
import pty
import termios

ctl = False


def child_prep(fd):
    global ctl
    if ctl:
        return
    ctl = True
    fcntl.ioctl(fd.fileno(), termios.TIOCSCTTY, 1)


def create_fds():
    ttys = []
    for i in range(10):
        (fd1, fd2) = pty.openpty()
        newattr = termios.tcgetattr(fd1)
        newattr[3] &= ~termios.ICANON & ~termios.ECHO
        termios.tcsetattr(fd1, termios.TCSADRAIN, newattr)
        ttys.append((os.fdopen(fd1, "wb"), os.fdopen(fd2, "rb")))
    return ttys


def filename(pipef):
    st = os.fstat(pipef.fileno())
    return 'tty[%x:%x]' % (st.st_rdev, st.st_dev)


def dump_opts(sockf):
    st = os.fstat(sockf.fileno())
    return "--external", 'tty[%x:%x]' % (st.st_rdev, st.st_dev)



import os
import tempfile, random


def mount(src, dst, shared, private, slave):
    cmd = "mount"
    if shared:
        cmd += " --make-shared"
    if private:
        cmd += " --make-private"
    if slave:
        cmd += " --make-slave"
    if src:
        cmd += " --bind '%s' '%s'" % (src, dst)
    else:
        cmd += " -t tmpfs none '%s'" % (dst)

    print(cmd)
    ret = os.system(cmd)
    if ret:
        print("failed")


root = tempfile.mkdtemp(prefix="root.mount", dir="/tmp")
mount(None, root, 1, 0, 0)
mounts = [root]

for i in range(10):
    dstdir = random.choice(mounts)
    dst = tempfile.mkdtemp(prefix="mount", dir=dstdir)
    src = random.choice(mounts + [None])
    mount(src, dst,
          random.randint(0, 100) > 50,
          random.randint(0, 100) > 90,
          random.randint(0, 100) > 50)
    mounts.append(dst)



import sys
import socket
import subprocess

def setup_swrk():
    print('Connecting to CRIU in swrk mode.')
    s1, s2 = socket.socketpair(socket.AF_UNIX, socket.SOCK_SEQPACKET)

    kwargs = {}
    if sys.version_info.major == 3:
        kwargs["pass_fds"] = [s1.fileno()]

    swrk = subprocess.Popen(['./criu', "swrk", "%d" % s1.fileno()], **kwargs)
    s1.close()
    return swrk, s2




#!/usr/bin/python
# Test criu errno

import socket, os, errno
import rpc_pb2 as rpc
import argparse

parser = argparse.ArgumentParser(description="Test errno reported by CRIU RPC")
parser.add_argument('socket', type=str, help="CRIU service socket")
parser.add_argument('dir',
                    type=str,
                    help="Directory where CRIU images should be placed")

args = vars(parser.parse_args())


# Prepare dir for images
class test:
    def __init__(self):
        self.imgs_fd = os.open(args['dir'], os.O_DIRECTORY)
        self.s = -1
        self._MAX_MSG_SIZE = 1024

    def connect(self):
        self.s = socket.socket(socket.AF_UNIX, socket.SOCK_SEQPACKET)
        self.s.connect(args['socket'])

    def get_base_req(self):
        req = rpc.criu_req()
        req.opts.log_level = 4
        req.opts.images_dir_fd = self.imgs_fd
        return req

    def send_req(self, req):
        self.connect()
        self.s.send(req.SerializeToString())

    def recv_resp(self):
        resp = rpc.criu_resp()
        resp.ParseFromString(self.s.recv(self._MAX_MSG_SIZE))
        return resp

    def check_resp(self, resp, typ, err):
        if resp.type != typ:
            raise Exception('Unexpected responce type ' + str(resp.type))

        if resp.success:
            raise Exception('Unexpected success = True')

        if err and resp.cr_errno != err:
            raise Exception('Unexpected cr_errno ' + str(resp.cr_errno))

    def no_process(self):
        print('Try to dump unexisting process')
        # Get pid of non-existing process.
        # Suppose max_pid is not taken by any process.
        with open("/proc/sys/kernel/pid_max", "r") as f:
            pid = int(f.readline())
            try:
                os.kill(pid, 0)
            except OSError:
                pass
            else:
                raise Exception('max pid is taken')

        # Ask criu to dump non-existing process.
        req = self.get_base_req()
        req.type = rpc.DUMP
        req.opts.pid = pid

        self.send_req(req)
        resp = self.recv_resp()

        self.check_resp(resp, rpc.DUMP, errno.ESRCH)

        print('Success')

    def process_exists(self):
        print(
            'Try to restore process which pid is already taken by other process'
        )

        # Perform self-dump
        req = self.get_base_req()
        req.type = rpc.DUMP
        req.opts.leave_running = True

        self.send_req(req)
        resp = self.recv_resp()

        if resp.success != True:
            raise Exception('Self-dump failed')

        # Ask to restore process from images of ourselves
        req = self.get_base_req()
        req.type = rpc.RESTORE

        self.send_req(req)
        resp = self.recv_resp()

        self.check_resp(resp, rpc.RESTORE, errno.EEXIST)

        print('Success')

    def bad_options(self):
        print('Try to send criu invalid opts')

        # Subdirs are not allowed in log_file
        req = self.get_base_req()
        req.type = rpc.DUMP
        req.opts.log_file = "../file.log"

        self.send_req(req)
        resp = self.recv_resp()

        self.check_resp(resp, rpc.DUMP, errno.EBADRQC)

        print('Success')

    def bad_request(self):
        print('Try to send criu invalid request type')

        req = self.get_base_req()
        req.type = rpc.NOTIFY

        self.send_req(req)
        resp = self.recv_resp()

        self.check_resp(resp, rpc.EMPTY, None)

        print('Success')

    def run(self):
        self.no_process()
        self.process_exists()
        self.bad_options()
        self.bad_request()


t = test()
t.run()



#!/usr/bin/env python

import sys
import os

actions = set(['pre-dump', 'pre-restore', 'post-dump', 'setup-namespaces', \
  'post-setup-namespaces', 'post-restore', 'post-resume', \
  'network-lock', 'network-unlock' ])
errors = []
af = os.path.dirname(os.path.abspath(__file__)) + '/actions_called.txt'

for act in open(af):
    act = act.strip().split()
    act.append('EMPTY')
    act.append('EMPTY')

    if act[0] == 'EMPTY':
        raise Exception("Error in test, bogus actions line")

    if act[1] == 'EMPTY':
        errors.append('Action %s misses CRTOOLS_IMAGE_DIR' % act[0])

    if act[0] in ('post-dump', 'setup-namespaces', 'post-setup-namespaces', \
      'post-restore', 'post-resume', 'network-lock', 'network-unlock'):
        if act[2] == 'EMPTY':
            errors.append('Action %s misses CRTOOLS_INIT_PID' % act[0])
        elif not act[2].isdigit() or int(act[2]) == 0:
            errors.append('Action %s PID is not number (%s)' %
                          (act[0], act[2]))

    actions -= set([act[0]])

if actions:
    errors.append('Not all actions called: %r' % actions)

if errors:
    for x in errors:
        print(x)
    sys.exit(1)

print('PASS')



import ctypes
import os
libc = ctypes.CDLL(None)


def memfd_create(name, flags):
    return libc.memfd_create(name.encode('utf8'), flags)


def create_fds():
    def create_memfd_pair(name):
        fd = memfd_create(name, 0)
        fw = open('/proc/self/fd/{}'.format(fd), 'wb')
        fr = open('/proc/self/fd/{}'.format(fd), 'rb')
        os.close(fd)
        return (fw, fr)

    return [create_memfd_pair("name{}".format(i)) for i in range(10)]


def filename(f):
    name = os.readlink('/proc/self/fd/{}'.format(f.fileno()))
    name = name.replace(' (deleted)', '')
    return name


def dump_opts(sockf):
    return []



#!/usr/bin/env python
import os, pty, sys, subprocess
import termios, fcntl, time

cr_bin = "../../../criu/criu"

os.chdir(os.getcwd())


def create_pty():
    (fd1, fd2) = pty.openpty()
    return (os.fdopen(fd1, "wb"), os.fdopen(fd2, "wb"))


if not os.access("work", os.X_OK):
    os.mkdir("work", 0o755)

open("running", "w").close()
m, s = create_pty()
p = os.pipe()
pr = os.fdopen(p[0], "r")
pw = os.fdopen(p[1], "w")

pid = os.fork()
if pid == 0:
    m.close()
    os.setsid()
    os.dup2(s.fileno(), 0)
    os.dup2(s.fileno(), 1)
    os.dup2(s.fileno(), 2)
    fcntl.ioctl(s.fileno(), termios.TIOCSCTTY, 1)
    pr.close()
    pw.close()
    while True:
        if not os.access("running", os.F_OK):
            sys.exit(0)
        time.sleep(1)
    sys.exit(1)

pw.close()
pr.read(1)

cmd = [cr_bin, "dump", "-j", "-t", str(pid), "-D", "work", "-v"]
print("Run: %s" % " ".join(cmd))
ret = subprocess.Popen(cmd).wait()
if ret != 0:
    sys.exit(1)
os.wait()

os.unlink("running")
m, s = create_pty()
cpid = os.fork()
if cpid == 0:
    os.setsid()
    fcntl.ioctl(m.fileno(), termios.TIOCSCTTY, 1)
    cmd = [cr_bin, "restore", "-j", "-D", "work", "-v"]
    print("Run: %s" % " ".join(cmd))
    ret = subprocess.Popen([cr_bin, "restore", "-j", "-D", "work",
                            "-v"]).wait()
    if ret != 0:
        sys.exit(1)
    sys.exit(0)

pid, status = os.wait()
if status != 0:
    print("A child process exited with %d" % status)
    sys.exit(1)



import os
import tempfile

id_str = ""


def create_fds():
    tdir = tempfile.mkdtemp("zdtm.inhfd.XXXXXX")
    if os.system("mount -t tmpfs zdtm.inhfd %s" % tdir) != 0:
        raise Exception("Unable to mount tmpfs")
    tfifo = os.path.join(tdir, "test_fifo")
    os.mkfifo(tfifo)
    fd2 = open(tfifo, "w+b", buffering=0)
    fd1 = open(tfifo, "rb")
    os.system("umount -l %s" % tdir)
    os.rmdir(tdir)

    mnt_id = -1
    with open("/proc/self/fdinfo/%d" % fd1.fileno()) as f:
        for line in f:
            line = line.split()
            if line[0] == "mnt_id:":
                mnt_id = int(line[1])
                break
        else:
            raise Exception("Unable to find mnt_id")

    global id_str
    id_str = "file[%x:%x]" % (mnt_id, os.fstat(fd1.fileno()).st_ino)

    return [(fd2, fd1)]


def filename(pipef):
    return id_str


def dump_opts(sockf):
    return ["--external", id_str]



# Define structures and constants for generating elf file.
import ctypes

Elf64_Half = ctypes.c_uint16  # typedef uint16_t Elf64_Half;
Elf64_Word = ctypes.c_uint32  # typedef uint32_t Elf64_Word;
Elf64_Addr = ctypes.c_uint64  # typedef uint64_t Elf64_Addr;
Elf64_Off = ctypes.c_uint64  # typedef uint64_t Elf64_Off;
Elf64_Xword = ctypes.c_uint64  # typedef uint64_t Elf64_Xword;

# Elf64_Ehdr related constants.

# e_ident size.
EI_NIDENT = 16  # #define EI_NIDENT (16)

EI_MAG0 = 0  # #define EI_MAG0         0               /* File identification byte 0 index */
ELFMAG0 = 0x7f  # #define ELFMAG0         0x7f            /* Magic number byte 0 */

EI_MAG1 = 1  # #define EI_MAG1         1               /* File identification byte 1 index */
ELFMAG1 = ord(
    'E')  # #define ELFMAG1         'E'             /* Magic number byte 1 */

EI_MAG2 = 2  # #define EI_MAG2         2               /* File identification byte 2 index */
ELFMAG2 = ord(
    'L')  # #define ELFMAG2         'L'             /* Magic number byte 2 */

EI_MAG3 = 3  # #define EI_MAG3         3               /* File identification byte 3 index */
ELFMAG3 = ord(
    'F')  # #define ELFMAG3         'F'             /* Magic number byte 3 */

EI_CLASS = 4  # #define EI_CLASS        4               /* File class byte index */

EI_DATA = 5  # #define EI_DATA         5               /* Data encoding byte index */

EI_VERSION = 6  # #define EI_VERSION      6               /* File version byte index */

ELFDATA2LSB = 1  # #define ELFDATA2LSB     1               /* 2's complement, little endian */

ELFCLASS64 = 2  # #define ELFCLASS64      2               /* 64-bit objects */

# Legal values for e_type (object file type).
ET_CORE = 4  # #define ET_CORE         4               /* Core file */

# Legal values for e_machine (architecture).
EM_X86_64 = 62  # #define EM_X86_64       62              /* AMD x86-64 architecture */

# Legal values for e_version (version).
EV_CURRENT = 1  # #define EV_CURRENT      1               /* Current version */


class Elf64_Ehdr(ctypes.Structure):  # typedef struct
    _fields_ = [  # {
        ("e_ident",
         ctypes.c_ubyte * EI_NIDENT),  #   unsigned char e_ident[EI_NIDENT];
        ("e_type", Elf64_Half),  #   Elf64_Half e_type;
        ("e_machine", Elf64_Half),  #   Elf64_Half e_machine;
        ("e_version", Elf64_Word),  #   Elf64_Word e_version;
        ("e_entry", Elf64_Addr),  #   Elf64_Addr e_entry;
        ("e_phoff", Elf64_Off),  #   Elf64_Off e_phoff;
        ("e_shoff", Elf64_Off),  #   Elf64_Off e_shoff;
        ("e_flags", Elf64_Word),  #   Elf64_Word e_flags;
        ("e_ehsize", Elf64_Half),  #   Elf64_Half e_ehsize;
        ("e_phentsize", Elf64_Half),  #   Elf64_Half e_phentsize;
        ("e_phnum", Elf64_Half),  #   Elf64_Half e_phnum;
        ("e_shentsize", Elf64_Half),  #   Elf64_Half e_shentsize;
        ("e_shnum", Elf64_Half),  #   Elf64_Half e_shnum;
        ("e_shstrndx", Elf64_Half)  #   Elf64_Half e_shstrndx;
    ]  # } Elf64_Ehdr;


# Elf64_Phdr related constants.

# Legal values for p_type (segment type).
PT_LOAD = 1  # #define PT_LOAD         1               /* Loadable program segment */
PT_NOTE = 4  # #define PT_NOTE         4               /* Auxiliary information */

# Legal values for p_flags (segment flags).
PF_X = 1  # #define PF_X            (1 << 0)        /* Segment is executable */
PF_W = 1 << 1  # #define PF_W            (1 << 1)        /* Segment is writable */
PF_R = 1 << 2  # #define PF_R            (1 << 2)        /* Segment is readable */


class Elf64_Phdr(ctypes.Structure):  # typedef struct
    _fields_ = [  # {
        ("p_type", Elf64_Word),  #   Elf64_Word p_type;
        ("p_flags", Elf64_Word),  #   Elf64_Word p_flags;
        ("p_offset", Elf64_Off),  #   Elf64_Off p_offset;
        ("p_vaddr", Elf64_Addr),  #   Elf64_Addr p_vaddr;
        ("p_paddr", Elf64_Addr),  #   Elf64_Addr p_paddr;
        ("p_filesz", Elf64_Xword),  #   Elf64_Xword p_filesz;
        ("p_memsz", Elf64_Xword),  #   Elf64_Xword p_memsz;
        ("p_align", Elf64_Xword),  #   Elf64_Xword p_align;
    ]  # } Elf64_Phdr;


# Elf64_auxv_t related constants.


class _Elf64_auxv_t_U(ctypes.Union):
    _fields_ = [("a_val", ctypes.c_uint64)]


class Elf64_auxv_t(ctypes.Structure):  # typedef struct
    _fields_ = [  # {
        ("a_type",
         ctypes.c_uint64),  #   uint64_t a_type;              /* Entry type */
        ("a_un", _Elf64_auxv_t_U)  #   union
        #     {
        #       uint64_t a_val;           /* Integer value */
        #       /* We use to have pointer elements added here.  We cannot do that,
        #          though, since it does not work when using 32-bit definitions
        #          on 64-bit platforms and vice versa.  */
        #     } a_un;
    ]  # } Elf64_auxv_t;


# Elf64_Nhdr related constants.

NT_PRSTATUS = 1  # #define NT_PRSTATUS     1               /* Contains copy of prstatus struct */
NT_FPREGSET = 2  # #define NT_FPREGSET     2               /* Contains copy of fpregset struct */
NT_PRPSINFO = 3  # #define NT_PRPSINFO     3               /* Contains copy of prpsinfo struct */
NT_AUXV = 6  # #define NT_AUXV         6               /* Contains copy of auxv array */
NT_SIGINFO = 0x53494749  # #define NT_SIGINFO      0x53494749      /* Contains copy of siginfo_t,
#                                         size might increase */
NT_FILE = 0x46494c45  # #define NT_FILE         0x46494c45      /* Contains information about mapped
#                                         files */
NT_X86_XSTATE = 0x202  # #define NT_X86_XSTATE   0x202           /* x86 extended state using xsave */


class Elf64_Nhdr(ctypes.Structure):  # typedef struct
    _fields_ = [  # {
        (
            "n_namesz", Elf64_Word
        ),  #   Elf64_Word n_namesz;                  /* Length of the note's name.  */
        (
            "n_descsz", Elf64_Word
        ),  #   Elf64_Word n_descsz;                  /* Length of the note's descriptor.  */
        ("n_type", Elf64_Word
         ),  #   Elf64_Word n_type;                    /* Type of the note.  */
    ]  # } Elf64_Nhdr;


# Elf64_Shdr related constants.


class Elf64_Shdr(ctypes.Structure):  # typedef struct
    _fields_ = [  # {
        (
            "sh_name", Elf64_Word
        ),  #   Elf64_Word    sh_name;                /* Section name (string tbl index) */
        ("sh_type", Elf64_Word
         ),  #   Elf64_Word    sh_type;                /* Section type */
        ("sh_flags", Elf64_Xword
         ),  #   Elf64_Xword   sh_flags;               /* Section flags */
        (
            "sh_addr", Elf64_Addr
        ),  #   Elf64_Addr    sh_addr;                /* Section virtual addr at execution */
        (
            "sh_offset", Elf64_Off
        ),  #   Elf64_Off     sh_offset;              /* Section file offset */
        (
            "sh_size", Elf64_Xword
        ),  #   Elf64_Xword   sh_size;                /* Section size in bytes */
        (
            "sh_link", Elf64_Word
        ),  #   Elf64_Word    sh_link;                /* Link to another section */
        (
            "sh_info", Elf64_Word
        ),  #   Elf64_Word    sh_info;                /* Additional section information */
        ("sh_addralign", Elf64_Xword
         ),  #   Elf64_Xword   sh_addralign;           /* Section alignment */
        (
            "sh_entsize", Elf64_Xword
        )  #   Elf64_Xword   sh_entsize;             /* Entry size if section holds table */
    ]  # } Elf64_Shdr;


# elf_prstatus related constants.


# Signal info.
class elf_siginfo(ctypes.Structure):  # struct elf_siginfo
    _fields_ = [  #   {
        ("si_signo", ctypes.c_int
         ),  #     int si_signo;                       /* Signal number.  */
        ("si_code", ctypes.c_int
         ),  #     int si_code;                        /* Extra code.  */
        ("si_errno", ctypes.c_int
         )  #     int si_errno;                       /* Errno.  */
    ]  #   };


# A time value that is accurate to the nearest
# microsecond but also has a range of years.
class timeval(ctypes.Structure):  # struct timeval
    _fields_ = [  #   {
        ("tv_sec",
         ctypes.c_long),  #     __time_t tv_sec;            /* Seconds.  */
        ("tv_usec", ctypes.c_long
         )  #     __suseconds_t tv_usec;      /* Microseconds.  */
    ]  #   };


class user_regs_struct(ctypes.Structure):  # struct user_regs_struct
    _fields_ = [  # {
        ("r15",
         ctypes.c_ulonglong),  #   __extension__ unsigned long long int r15;
        ("r14",
         ctypes.c_ulonglong),  #   __extension__ unsigned long long int r14;
        ("r13",
         ctypes.c_ulonglong),  #   __extension__ unsigned long long int r13;
        ("r12",
         ctypes.c_ulonglong),  #   __extension__ unsigned long long int r12;
        ("rbp",
         ctypes.c_ulonglong),  #   __extension__ unsigned long long int rbp;
        ("rbx",
         ctypes.c_ulonglong),  #   __extension__ unsigned long long int rbx;
        ("r11",
         ctypes.c_ulonglong),  #   __extension__ unsigned long long int r11;
        ("r10",
         ctypes.c_ulonglong),  #   __extension__ unsigned long long int r10;
        ("r9",
         ctypes.c_ulonglong),  #   __extension__ unsigned long long int r9;
        ("r8",
         ctypes.c_ulonglong),  #   __extension__ unsigned long long int r8;
        ("rax",
         ctypes.c_ulonglong),  #   __extension__ unsigned long long int rax;
        ("rcx",
         ctypes.c_ulonglong),  #   __extension__ unsigned long long int rcx;
        ("rdx",
         ctypes.c_ulonglong),  #   __extension__ unsigned long long int rdx;
        ("rsi",
         ctypes.c_ulonglong),  #   __extension__ unsigned long long int rsi;
        ("rdi",
         ctypes.c_ulonglong),  #   __extension__ unsigned long long int rdi;
        ("orig_rax", ctypes.c_ulonglong
         ),  #   __extension__ unsigned long long int orig_rax;
        ("rip",
         ctypes.c_ulonglong),  #   __extension__ unsigned long long int rip;
        ("cs",
         ctypes.c_ulonglong),  #   __extension__ unsigned long long int cs;
        ("eflags",
         ctypes.c_ulonglong),  #   __extension__ unsigned long long int eflags;
        ("rsp",
         ctypes.c_ulonglong),  #   __extension__ unsigned long long int rsp;
        ("ss",
         ctypes.c_ulonglong),  #   __extension__ unsigned long long int ss;
        ("fs_base", ctypes.c_ulonglong
         ),  #   __extension__ unsigned long long int fs_base;
        ("gs_base", ctypes.c_ulonglong
         ),  #   __extension__ unsigned long long int gs_base;
        ("ds",
         ctypes.c_ulonglong),  #   __extension__ unsigned long long int ds;
        ("es",
         ctypes.c_ulonglong),  #   __extension__ unsigned long long int es;
        ("fs",
         ctypes.c_ulonglong),  #   __extension__ unsigned long long int fs;
        ("gs", ctypes.c_ulonglong
         )  #   __extension__ unsigned long long int gs;
    ]  # };


#elf_greg_t    = ctypes.c_ulonglong
#ELF_NGREG = ctypes.sizeof(user_regs_struct)/ctypes.sizeof(elf_greg_t)
#elf_gregset_t = elf_greg_t*ELF_NGREG
elf_gregset_t = user_regs_struct


class elf_prstatus(ctypes.Structure):  # struct elf_prstatus
    _fields_ = [  #   {
        (
            "pr_info", elf_siginfo
        ),  #     struct elf_siginfo pr_info;         /* Info associated with signal.  */
        ("pr_cursig", ctypes.c_short
         ),  #     short int pr_cursig;                /* Current signal.  */
        (
            "pr_sigpend", ctypes.c_ulong
        ),  #     unsigned long int pr_sigpend;       /* Set of pending signals.  */
        (
            "pr_sighold", ctypes.c_ulong
        ),  #     unsigned long int pr_sighold;       /* Set of held signals.  */
        ("pr_pid", ctypes.c_int),  #     __pid_t pr_pid;
        ("pr_ppid", ctypes.c_int),  #     __pid_t pr_ppid;
        ("pr_pgrp", ctypes.c_int),  #     __pid_t pr_pgrp;
        ("pr_sid", ctypes.c_int),  #     __pid_t pr_sid;
        ("pr_utime",
         timeval),  #     struct timeval pr_utime;            /* User time.  */
        ("pr_stime", timeval
         ),  #     struct timeval pr_stime;            /* System time.  */
        (
            "pr_cutime", timeval
        ),  #     struct timeval pr_cutime;           /* Cumulative user time.  */
        (
            "pr_cstime", timeval
        ),  #     struct timeval pr_cstime;           /* Cumulative system time.  */
        ("pr_reg", elf_gregset_t
         ),  #     elf_gregset_t pr_reg;               /* GP registers.  */
        (
            "pr_fpvalid", ctypes.c_int
        )  #     int pr_fpvalid;                     /* True if math copro being used.  */
    ]  #   };


# elf_prpsinfo related constants.

ELF_PRARGSZ = 80  # #define ELF_PRARGSZ     (80)    /* Number of chars for args.  */


class elf_prpsinfo(ctypes.Structure):  # struct elf_prpsinfo
    _fields_ = [  #   {
        (
            "pr_state", ctypes.c_byte
        ),  #     char pr_state;                      /* Numeric process state.  */
        (
            "pr_sname", ctypes.c_char
        ),  #     char pr_sname;                      /* Char for pr_state.  */
        ("pr_zomb", ctypes.c_byte
         ),  #     char pr_zomb;                       /* Zombie.  */
        ("pr_nice", ctypes.c_byte
         ),  #     char pr_nice;                       /* Nice val.  */
        ("pr_flag", ctypes.c_ulong
         ),  #     unsigned long int pr_flag;          /* Flags.  */
        # #if __WORDSIZE == 32
        #     unsigned short int pr_uid;
        #     unsigned short int pr_gid;
        # #else
        ("pr_uid", ctypes.c_uint),  #     unsigned int pr_uid;
        ("pr_gid", ctypes.c_uint),  #     unsigned int pr_gid;
        # #endif
        ("pr_pid", ctypes.c_int),  #     int pr_pid, pr_ppid, pr_pgrp, pr_sid;
        ("pr_ppid", ctypes.c_int),
        ("pr_pgrp", ctypes.c_int),
        ("pr_sid", ctypes.c_int),
        #     /* Lots missing */
        (
            "pr_fname", ctypes.c_char * 16
        ),  #     char pr_fname[16];                  /* Filename of executable.  */
        (
            "pr_psargs", ctypes.c_char * ELF_PRARGSZ
        )  #     char pr_psargs[ELF_PRARGSZ];        /* Initial part of arg list.  */
    ]  #   };


class user_fpregs_struct(ctypes.Structure):  # struct user_fpregs_struct
    _fields_ = [  # {
        ("cwd", ctypes.c_ushort),  #   unsigned short int    cwd;
        ("swd", ctypes.c_ushort),  #   unsigned short int    swd;
        ("ftw", ctypes.c_ushort),  #   unsigned short int    ftw;
        ("fop", ctypes.c_ushort),  #   unsigned short int    fop;
        ("rip",
         ctypes.c_ulonglong),  #   __extension__ unsigned long long int rip;
        ("rdp",
         ctypes.c_ulonglong),  #   __extension__ unsigned long long int rdp;
        ("mxcsr", ctypes.c_uint),  #   unsigned int          mxcsr;
        ("mxcr_mask", ctypes.c_uint),  #   unsigned int          mxcr_mask;
        (
            "st_space", ctypes.c_uint * 32
        ),  #   unsigned int          st_space[32];   /* 8*16 bytes for each FP-reg = 128 bytes */
        (
            "xmm_space", ctypes.c_uint * 64
        ),  #   unsigned int          xmm_space[64];  /* 16*16 bytes for each XMM-reg = 256 bytes */
        ("padding",
         ctypes.c_uint * 24),  #   unsigned int          padding[24];
    ]  # };


elf_fpregset_t = user_fpregs_struct

# siginfo_t related constants.

_SI_MAX_SIZE = 128
_SI_PAD_SIZE = (_SI_MAX_SIZE / ctypes.sizeof(ctypes.c_int)) - 4


#          /* kill().  */
class _siginfo_t_U_kill(ctypes.Structure):  #         struct
    _fields_ = [  #           {
        ("si_pid", ctypes.c_int
         ),  #             __pid_t si_pid;     /* Sending process ID.  */
        (
            "si_uid", ctypes.c_uint
        )  #             __uid_t si_uid;     /* Real user ID of sending process.  */
    ]  #           } _kill;


# Type for data associated with a signal.
class sigval_t(ctypes.Union):  # typedef union sigval
    _fields_ = [  #   {
        ("sival_int", ctypes.c_int),  #     int sival_int;
        ("sical_ptr", ctypes.c_void_p),  #     void *sival_ptr;
    ]  #   } sigval_t;


    #         /* POSIX.1b timers.  */
class _siginfo_t_U_timer(ctypes.Structure):  #         struct
    _fields_ = [  #           {
        ("si_tid",
         ctypes.c_int),  #             int si_tid;         /* Timer ID.  */
        ("si_overrun", ctypes.c_int
         ),  #             int si_overrun;     /* Overrun count.  */
        ("si_sigval", sigval_t
         )  #             sigval_t si_sigval; /* Signal value.  */
    ]  #           } _timer;


    #         /* POSIX.1b signals.  */
class _siginfo_t_U_rt(ctypes.Structure):  #         struct
    _fields_ = [  #           {
        ("si_pid", ctypes.c_int
         ),  #             __pid_t si_pid;     /* Sending process ID.  */
        (
            "si_uid", ctypes.c_uint
        ),  #             __uid_t si_uid;     /* Real user ID of sending process.  */
        ("si_sigval", sigval_t
         )  #             sigval_t si_sigval; /* Signal value.  */
    ]  #           } _rt;


    #         /* SIGCHLD.  */
class _siginfo_t_U_sigchld(ctypes.Structure):  #         struct
    _fields_ = [  #           {
        ("si_pid",
         ctypes.c_int),  #             __pid_t si_pid;     /* Which child.  */
        (
            "si_uid", ctypes.c_uint
        ),  #             __uid_t si_uid;     /* Real user ID of sending process.  */
        ("si_status", ctypes.c_int
         ),  #             int si_status;      /* Exit value or signal.  */
        ("si_utime", ctypes.c_long),  #             __sigchld_clock_t si_utime;
        ("si_stime", ctypes.c_long)  #             __sigchld_clock_t si_stime;
    ]  #           } _sigchld;


    #         /* SIGILL, SIGFPE, SIGSEGV, SIGBUS.  */
class _siginfo_t_U_sigfault(ctypes.Structure):  #         struct
    _fields_ = [  #           {
        ("si_addr", ctypes.c_void_p
         ),  #             void *si_addr;      /* Faulting insn/memory ref.  */
        (
            "si_addr_lsb", ctypes.c_short
        )  #             short int si_addr_lsb;      /* Valid LSB of the reported address.  */
    ]  #           } _sigfault;


    #         /* SIGPOLL.  */
class _siginfo_t_U_sigpoll(ctypes.Structure):  #         struct
    _fields_ = [  #           {
        ("si_band", ctypes.c_long
         ),  #             long int si_band;   /* Band event for SIGPOLL.  */
        ("si_fd", ctypes.c_int)  #             int si_fd;
    ]  #           } _sigpoll;


    #             /* SIGSYS.  */
class _siginfo_t_U_sigsys(ctypes.Structure):  #         struct
    _fields_ = [  #           {
        ("_call_addr", ctypes.c_void_p
         ),  #             void *_call_addr;   /* Calling user insn.  */
        (
            "_syscall", ctypes.c_int
        ),  #             int _syscall;       /* Triggering system call number.  */
        ("_arch", ctypes.c_uint
         )  #             unsigned int _arch; /* AUDIT_ARCH_* of syscall.  */
    ]  #           } _sigsys;


class _siginfo_t_U(ctypes.Union):  #     union
    _fields_ = [  #       {
        ("_pad",
         ctypes.c_int * _SI_PAD_SIZE),  #         int _pad[__SI_PAD_SIZE];
        #
        #          /* kill().  */
        ("_kill", _siginfo_t_U_kill),  #         struct
        #           {
        #             __pid_t si_pid;     /* Sending process ID.  */
        #             __uid_t si_uid;     /* Real user ID of sending process.  */
        #           } _kill;
        #
        #         /* POSIX.1b timers.  */
        ("_timer", _siginfo_t_U_timer),  #         struct
        #           {
        #             int si_tid;         /* Timer ID.  */
        #             int si_overrun;     /* Overrun count.  */
        #             sigval_t si_sigval; /* Signal value.  */
        #           } _timer;
        #
        #         /* POSIX.1b signals.  */
        ("_rt", _siginfo_t_U_rt),  #         struct
        #           {
        #             __pid_t si_pid;     /* Sending process ID.  */
        #             __uid_t si_uid;     /* Real user ID of sending process.  */
        #             sigval_t si_sigval; /* Signal value.  */
        #           } _rt;
        #
        #         /* SIGCHLD.  */
        ("_sigchld", _siginfo_t_U_sigchld),  #         struct
        #           {
        #             __pid_t si_pid;     /* Which child.  */
        #             __uid_t si_uid;     /* Real user ID of sending process.  */
        #             int si_status;      /* Exit value or signal.  */
        #             __sigchld_clock_t si_utime;
        #             __sigchld_clock_t si_stime;
        #           } _sigchld;
        #
        #         /* SIGILL, SIGFPE, SIGSEGV, SIGBUS.  */
        ("_sigfault", _siginfo_t_U_sigfault),  #         struct
        #           {
        #             void *si_addr;      /* Faulting insn/memory ref.  */
        #             short int si_addr_lsb;      /* Valid LSB of the reported address.  */
        #           } _sigfault;
        #
        #         /* SIGPOLL.  */
        ("_sigpoll", _siginfo_t_U_sigpoll),  #         struct
        #           {
        #             long int si_band;   /* Band event for SIGPOLL.  */
        #             int si_fd;
        #           } _sigpoll;
        #
        #             /* SIGSYS.  */
        ("_sigsys", _siginfo_t_U_sigpoll)  #         struct
        #           {
        #             void *_call_addr;   /* Calling user insn.  */
        #             int _syscall;       /* Triggering system call number.  */
        #             unsigned int _arch; /* AUDIT_ARCH_* of syscall.  */
        #           } _sigsys;
    ]  #       } _sifields;


class siginfo_t(ctypes.Structure):  # typedef struct
    _fields_ = [  #   {
        ("si_signo", ctypes.c_int
         ),  #     int si_signo;               /* Signal number.  */
        (
            "si_errno", ctypes.c_int
        ),  #     int si_errno;               /* If non-zero, an errno value associated with
        #                                    this signal, as defined in <errno.h>.  */
        ("si_code", ctypes.c_int
         ),  #     int si_code;                /* Signal code.  */
        #
        ("_sifields", _siginfo_t_U)  #     union
        #       {
        #         int _pad[__SI_PAD_SIZE];
        #
        #          /* kill().  */
        #         struct
        #           {
        #             __pid_t si_pid;     /* Sending process ID.  */
        #             __uid_t si_uid;     /* Real user ID of sending process.  */
        #           } _kill;
        #
        #         /* POSIX.1b timers.  */
        #         struct
        #           {
        #             int si_tid;         /* Timer ID.  */
        #             int si_overrun;     /* Overrun count.  */
        #             sigval_t si_sigval; /* Signal value.  */
        #           } _timer;
        #
        #         /* POSIX.1b signals.  */
        #         struct
        #           {
        #             __pid_t si_pid;     /* Sending process ID.  */
        #             __uid_t si_uid;     /* Real user ID of sending process.  */
        #             sigval_t si_sigval; /* Signal value.  */
        #           } _rt;
        #
        #         /* SIGCHLD.  */
        #         struct
        #           {
        #             __pid_t si_pid;     /* Which child.  */
        #             __uid_t si_uid;     /* Real user ID of sending process.  */
        #             int si_status;      /* Exit value or signal.  */
        #             __sigchld_clock_t si_utime;
        #             __sigchld_clock_t si_stime;
        #           } _sigchld;
        #
        #         /* SIGILL, SIGFPE, SIGSEGV, SIGBUS.  */
        #         struct
        #           {
        #             void *si_addr;      /* Faulting insn/memory ref.  */
        #             short int si_addr_lsb;      /* Valid LSB of the reported address.  */
        #           } _sigfault;
        #
        #         /* SIGPOLL.  */
        #         struct
        #           {
        #             long int si_band;   /* Band event for SIGPOLL.  */
        #             int si_fd;
        #           } _sigpoll;
        #
        #             /* SIGSYS.  */
        #         struct
        #           {
        #             void *_call_addr;   /* Calling user insn.  */
        #             int _syscall;       /* Triggering system call number.  */
        #             unsigned int _arch; /* AUDIT_ARCH_* of syscall.  */
        #           } _sigsys;
        #       } _sifields;
    ]  #   } siginfo_t __SI_ALIGNMENT;


# xsave related.


class ymmh_struct(ctypes.Structure):  # struct ymmh_struct {
    _fields_ = [("ymmh_space", 64 * ctypes.c_uint
                 )  #         u32                             ymmh_space[64];
                ]  # } __packed;


class xsave_hdr_struct(ctypes.Structure):  # struct xsave_hdr_struct {
    _fields_ = [
        ("xstate_bv", ctypes.c_ulonglong
         ),  #         u64                             xstate_bv;
        ("reserved1", ctypes.c_ulonglong *
         2),  #         u64                             reserved1[2];
        ("reserved2", ctypes.c_ulonglong * 5
         )  #         u64                             reserved2[5];
    ]  # } __packed;


class i387_fxsave_struct(ctypes.Structure):  # struct i387_fxsave_struct {
    _fields_ = [
        (
            "cwd", ctypes.c_ushort
        ),  #         u16                             cwd; /* Control Word                    */
        (
            "swd", ctypes.c_ushort
        ),  #         u16                             swd; /* Status Word                     */
        (
            "twd", ctypes.c_ushort
        ),  #         u16                             twd; /* Tag Word                        */
        (
            "fop", ctypes.c_ushort
        ),  #         u16                             fop; /* Last Instruction Opcode         */
        #         union {
        #                 struct {
        (
            "rip", ctypes.c_ulonglong
        ),  #                         u64             rip; /* Instruction Pointer             */
        (
            "rdp", ctypes.c_ulonglong
        ),  #                         u64             rdp; /* Data Pointer                    */
        #                 };
        #                 struct {
        #                         u32             fip; /* FPU IP Offset                   */
        #                         u32             fcs; /* FPU IP Selector                 */
        #                         u32             foo; /* FPU Operand Offset              */
        #                         u32             fos; /* FPU Operand Selector            */
        #                 };
        #         };
        (
            "mxcsr", ctypes.c_uint
        ),  #         u32                             mxcsr;          /* MXCSR Register State */
        (
            "mxcsr_mask", ctypes.c_uint
        ),  #         u32                             mxcsr_mask;     /* MXCSR Mask           */
        #
        #         /* 8*16 bytes for each FP-reg = 128 bytes                               */
        ("st_space", ctypes.c_uint * 32
         ),  #         u32                             st_space[32];
        #
        #         /* 16*16 bytes for each XMM-reg = 256 bytes                             */
        ("xmm_space", ctypes.c_uint * 64
         ),  #         u32                             xmm_space[64];
        #
        ("padding", ctypes.c_uint * 12
         ),  #         u32                             padding[12];
        #
        #         union {
        ("padding1", ctypes.c_uint * 12
         )  #                 u32                     padding1[12];
        #                 u32                     sw_reserved[12];
        #         };
        #
    ]  # } __aligned(16);


class elf_xsave_struct(ctypes.Structure):  # struct xsave_struct {
    _fields_ = [
        ("i387",
         i387_fxsave_struct),  #         struct i387_fxsave_struct       i387;
        ("xsave_hdr", xsave_hdr_struct
         ),  #         struct xsave_hdr_struct         xsave_hdr;
        ("ymmh", ymmh_struct)  #         struct ymmh_struct              ymmh;
    ]  # } __aligned(FP_MIN_ALIGN_BYTES) __packed;



import os


def create_fds():
    pipes = []
    for i in range(10):
        (fd1, fd2) = os.pipe()
        pipes.append((os.fdopen(fd2, "wb"), os.fdopen(fd1, "rb")))
    return pipes


def filename(pipef):
    return 'pipe:[%d]' % os.fstat(pipef.fileno()).st_ino


def dump_opts(sockf):
    return []



#!/usr/bin/env python

import sys, os
import hashlib
from subprocess import Popen, PIPE

str2 = "test_test" * (1 << 20)
str1 = "Test_Test!"

src = os.getenv("TCP_SRC", "127.0.0.1")
dst = os.getenv("TCP_DST", "127.0.0.1")
sport = os.getenv("TCP_SPORT", "12345")
dport = os.getenv("TCP_DPORT", "54321")

print(sys.argv[1])
args = [
    sys.argv[1], "--addr", src, "--port", sport, "--seq", "555", "--next",
    "--addr", dst, "--port", dport, "--seq", "666", "--reverse", "--",
    "./tcp-test.py"
]

p1 = Popen(args + ["dst"], stdout=PIPE, stdin=PIPE)

args.remove("--reverse")

p2 = Popen(args + ["src"], stdout=PIPE, stdin=PIPE)

p1.stdout.read(5)
p2.stdout.read(5)
p1.stdin.write("start")
p2.stdin.write("start")

p1.stdin.write(str1)
p1.stdin.close()
p2.stdin.write(str2)
p2.stdin.close()

s = p1.stdout.read()
m = hashlib.md5()
m.update(str2)
str2 = m.hexdigest()

if str2 != eval(s):
    print("FAIL", repr(str2), repr(s))
    sys.exit(5)

s = p1.stdout.read()
m = hashlib.md5()
m.update(str1)
str1 = m.hexdigest()

s = p2.stdout.read()
if str1 != eval(s):
    print("FAIL", repr(str1), s)
    sys.exit(5)

if p1.wait():
    sys.exit(1)
if p2.wait():
    sys.exit(1)

print("PASS")



#!/usr/bin/python

import socket, os, sys, errno
import rpc_pb2 as rpc
import argparse

parser = argparse.ArgumentParser(description="Test page-server using CRIU RPC")
parser.add_argument('socket', type=str, help="CRIU service socket")
parser.add_argument('dir',
                    type=str,
                    help="Directory where CRIU images should be placed")

args = vars(parser.parse_args())

# Connect to service socket
s = socket.socket(socket.AF_UNIX, socket.SOCK_SEQPACKET)
s.connect(args['socket'])

# Start page-server
print('Starting page-server')
req = rpc.criu_req()
req.type = rpc.PAGE_SERVER
req.opts.log_file = 'page-server.log'
req.opts.log_level = 4
req.opts.images_dir_fd = os.open(args['dir'], os.O_DIRECTORY)

s.send(req.SerializeToString())

resp = rpc.criu_resp()
MAX_MSG_SIZE = 1024
resp.ParseFromString(s.recv(MAX_MSG_SIZE))

if resp.type != rpc.PAGE_SERVER:
    print('Unexpected msg type')
    sys.exit(1)
else:
    if resp.success:
        # check if pid even exists
        try:
            os.kill(resp.ps.pid, 0)
        except OSError as err:
            if err.errno == errno.ESRCH:
                print('No process with page-server pid %d' % (resp.ps.pid))
            else:
                print('Can\'t check that process %d exists' % (resp.ps.pid))
                sys.exit(1)
        print('Success, page-server pid %d started on port %u' %
              (resp.ps.pid, resp.ps.port))
    else:
        print('Failed to start page-server')
        sys.exit(1)

# Perform self-dump
print('Dumping myself using page-server')
req.type = rpc.DUMP
req.opts.ps.port = resp.ps.port
req.opts.ps.address = "127.0.0.1"
req.opts.log_file = 'dump.log'
req.opts.leave_running = True

s.close()
s = socket.socket(socket.AF_UNIX, socket.SOCK_SEQPACKET)
s.connect(args['socket'])
s.send(req.SerializeToString())

resp.ParseFromString(s.recv(MAX_MSG_SIZE))

if resp.type != rpc.DUMP:
    print('Unexpected msg type')
    sys.exit(1)
else:
    if resp.success:
        print('Success')
    else:
        print('Fail')
        sys.exit(1)
